/*
 * Copyright 2020 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"; you may not use
 * this file except in compliance with the License. You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.physical.pull.operators;

import io.confluent.ksql.execution.streams.materialization.Locator.KsqlPartitionLocation;
import java.util.List;

/**
 * Represents the leaf operators of physical plans that scan the data sources
 * (for example data stores).  The methods allow for dynamically getting/setting the partition
 * information that is only available at runtime.
 */
public interface DataSourceOperator {

  List<KsqlPartitionLocation> getPartitionLocations();

  void setPartitionLocations(List<KsqlPartitionLocation> partitionLocations);

  long getReturnedRowCount();
}
